#ifndef DUNE_TECTONIC_GLOBAL_NONLINEARITY_HH
#define DUNE_TECTONIC_GLOBAL_NONLINEARITY_HH

#include <dune/common/fmatrix.hh>
#include <dune/common/fvector.hh>
#include <dune/common/shared_ptr.hh>
#include <dune/istl/bcrsmatrix.hh>
#include <dune/istl/bvector.hh>
#include <dune/istl/matrixindexset.hh>

#include "frictionpotential.hh"
#include "localfriction.hh"

namespace Dune {
template <class MatrixTypeTEMPLATE, class VectorTypeTEMPLATE>
class GlobalNonlinearity {
protected:
  using SingletonVectorType = BlockVector<FieldVector<double, 1>>;

public:
  using IndexSet = Dune::MatrixIndexSet;
  using MatrixType = MatrixTypeTEMPLATE;
  using VectorType = VectorTypeTEMPLATE;
  using LocalMatrixType = typename MatrixType::block_type;
  using LocalVectorType = typename VectorType::block_type;
  static const int block_size = LocalVectorType::dimension;
  using FrictionType = LocalFriction<block_size>;

  double operator()(VectorType const &x) const {
    double tmp = 0;
    for (size_t i = 0; i < x.size(); ++i) {
      auto const res = restriction(i);
      tmp += (*res)(x[i]);
    }
    return tmp;
  }

  /*
    Return a restriction of the outer function to the i'th node.
  */
  virtual shared_ptr<LocalFriction<block_size>> restriction(int i) const = 0;

  void addHessian(VectorType const &v, MatrixType &hessian) const {
    for (size_t i = 0; i < v.size(); ++i) {
      auto const res = restriction(i);
      res->addHessian(v[i], hessian[i][i]);
    }
  }

  void directionalDomain(VectorType const &, VectorType const &,
                         Interval<double> &dom) const {
    dom[0] = -std::numeric_limits<double>::max();
    dom[1] = std::numeric_limits<double>::max();
  }

  void directionalSubDiff(VectorType const &u, VectorType const &v,
                          Interval<double> &subdifferential) const {
    subdifferential[0] = subdifferential[1] = 0;
    for (size_t i = 0; i < u.size(); ++i) {
      Interval<double> D;
      auto const res = restriction(i);
      res->directionalSubDiff(u[i], v[i], D);
      subdifferential[0] += D[0];
      subdifferential[1] += D[1];
    }
  }

  void addHessianIndices(Dune::MatrixIndexSet &indices) const {
    for (size_t i = 0; i < indices.rows(); ++i)
      indices.add(i, i);
  }

  void addGradient(VectorType const &v, VectorType &gradient) const {
    for (size_t i = 0; i < v.size(); ++i) {
      auto const res = restriction(i);
      res->addGradient(v[i], gradient[i]);
    }
  }

  double regularity(int i, typename VectorType::block_type const &x) const {
    auto const res = restriction(i);
    return res->regularity(x);
  }

  virtual void updateState(SingletonVectorType const &state) = 0;
};
}
#endif
